package net;

import game.Game;
import game.GameDynamics;

import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.net.Socket;
import java.util.Vector;

import util.Array;


public class GameServer implements Runnable, ChatListener{
  
  
  public String toString(){
    String result = ""+getNumUsers();
    for(int i=0;i<getNumUsers();i++){
      Agent a = getUser(i);
      if (a!=null){
        result+=" "+a.toString();
      }
    }
    return result;
  }
  Vector<Agent> users;
  int userLimit;
  GameDynamics dynamics;
  Game g;
  boolean gameInProgress;
  boolean lastRound;
  long currentGameIndex;
  static final String logSuffix = ".log";
  
  public static long DEFAULT_RECEIVING_TIMEOUT=60000;
  private static final int DEFAULT_NUM_ROUNDS=4;
 
  public long receivingTimeout=60000;
  private int numRounds=4;
 
  PrintStream log;
  
  boolean handshakeComplete;
  
  public GameServer(Game g, long currentGameIndex, String logLocation, int numRounds, long receivingTimeout){
    this.g = g;
    this.currentGameIndex = currentGameIndex;
    this.numRounds = numRounds;
    this.receivingTimeout=receivingTimeout;
    observedActions = new int[g.getNumPlayers()];
    outcomesSent = new int[g.getNumPlayers()];
    users = new Vector<Agent>();
    String logName = logLocation + currentGameIndex + logSuffix;
    try{
    log = new PrintStream(new FileOutputStream(logName));
    } catch (FileNotFoundException fnf){
      System.err.println("Log not started for game "+currentGameIndex);
      log = null;
    }
    gameInProgress=false;
    handshakeComplete=false;
  }
  public GameServer(Game g, long currentGameIndex, String logLocation){
    this(g, currentGameIndex,logLocation,DEFAULT_NUM_ROUNDS, DEFAULT_RECEIVING_TIMEOUT);
  }

  public void recordInLog(String line){
    if (log!=null){
      log.println(System.currentTimeMillis()+":"+line);
      log.flush();
    }
  }
  
  public void closeLog(){
    if (log!=null){
      log.flush();
      log.close();
    }
    log=null;
  }
  public void start(){
    Thread t = new Thread(this);
    t.start();
  }
  
  
  
  
  public boolean enoughUsers(){
    return getNumUsers()>=g.getNumPlayers();
  }
  
  
  /**
   * This is synchronized because I don't know what happens if
   * I am adding someone and then I test the size of the vector.
   * @param s
   */
  public synchronized void addClient(Socket s){
    recordInLog("Adding user:"+(users.size())+" IP:"+s.getInetAddress());
    users.add(new Agent(this,s));
  }
  
  public synchronized int getNumUsers(){
    return users.size();
  }
  
  
  
  private synchronized void removeUser(Agent a){
    System.err.println("Removing user");
    a.gameOver = true;
    users.remove(a);
    System.err.println("Users remaining:"+users.size());
  }
  
  
  public synchronized Agent getUser(int i){
    if (i<users.size()){
      return users.get(i);
    }
    return null;
  }

  
  
  public void abortGame(){
    while(getNumUsers()!=0){
      Agent agent = this.getUser(0);
      if (agent!=null){
        try {
          agent.sendMessage("ABORT");
        } catch (Exception e) {
          System.err.println("Error sending ABORT message to agent");
          e.printStackTrace();
        }
        removeUser(agent);
      }
    }
    recordInLog("game aborted");
    closeLog();
  }
  
  public void removeAllUsers(){
    while(getNumUsers()!=0){
      Agent agent = this.getUser(0);
      if (agent!=null){
        this.removeUser(agent);
      }
    }
  }
  
  private final int[] observedActions;
  private final int[] outcomesSent;
  
  public void resetActions(){
    for(int j=0;j<g.getNumPlayers();j++){
      observedActions[j]=-1;
    }
  }
  
  public void getActions() throws IOException{
    for(int j=0;j<g.getNumPlayers();j++){
      if (observedActions[j]==-1){
      Agent agent = this.getUser(j);
      if (agent==null || agent.gameOver){
        throw new IOException("Agent left the game");
      }
      observedActions[j] = agent.getActionObserved();
    }
    }
  }

  public boolean areActionsLegal(){
    return g.areLegalActions(observedActions);
  }
  
  public boolean playerAborted(){
    for(int i=0;i<observedActions.length;i++){
      if (observedActions[i]==g.getNumActions()){
        return true;
      }
    }
    return false;
  }

  public boolean actionsComplete(){
    for(int i=0;i<observedActions.length;i++){
      if (observedActions[i]==-1){
        return false;
      }
    }
    return true;
  }
  
  
  public void sendOutcomes() throws IOException{
    recordInLog("Actions:"+Array.arrayToNakedString(observedActions));
    for(int j=0;j<g.getNumPlayers();j++){
      Agent agent = this.getUser(j);
      if (agent==null || (agent.gameOver) || (this.getNumUsers()!=g.getNumPlayers())){
        throw new IOException("Some agent left the game");
      }
      System.arraycopy(observedActions,0,outcomesSent,0,observedActions.length);
      outcomesSent[0]=observedActions[j];
      outcomesSent[j]=observedActions[0];
      agent.observeOutcome(outcomesSent);
    }
    
  }
  
  @Override
  public void run() {
    try{
    System.out.println("Game begins!");
    lastRound = false;
    gameInProgress=true;
    
    
    
    
    for(int i=0;i<g.getNumPlayers();i++){
      Agent agent = this.getUser(i);
      if (agent!=null){
        try{
        agent.sendMessage("PING");
        } catch (IOException io){
          agent.gameOver = true;
          throw new IOException("Agent left the game before it began");
        }
      }
    }
    

    long currentTime = System.currentTimeMillis();
    long timeout = currentTime+receivingTimeout;
    boolean[] playerAcknowledged = new boolean[g.getNumPlayers()];
    for(int i=0;i<playerAcknowledged.length;i++){
      playerAcknowledged[i]=false;
    }
    int numAcknowledged=0;
    while((numAcknowledged<g.getNumPlayers()) && (System.currentTimeMillis()<timeout)){
    for(int i=0;i<g.getNumPlayers();i++){
      Agent agent = this.getUser(i);
      if (agent!=null && !playerAcknowledged[i]){
        try{
        String message = agent.peekAtMessage();
        if (message!=null){
          message = agent.receiveMessage();
          if (!"ACK".equals(message)){
            throw new IOException("ACK not received");
          }
          playerAcknowledged[i]=true;
          numAcknowledged++;
        }
        } catch (IOException io){
          agent.gameOver = true;
          throw new IOException("Agent left the game before it began");
        }
      }
    }
      Thread.sleep(100);
    }
    
    if (numAcknowledged!=g.getNumPlayers()){
      throw new IOException("Agent timed out during handshake");
      
    }
    for(int i=0;i<g.getNumPlayers();i++){
      Agent agent = this.getUser(i);
      if (agent!=null){
        try{
          agent.sendMessage("BEGIN:"+this.currentGameIndex+" LEMONADE "+g.getNumPlayers()+" "+g.getNumActions()+" "+i);
        } catch (IOException io){
          agent.gameOver = true;
          throw new IOException("Agent left the game before it began");
        }
      }
    }
    
    handshakeComplete = true;

    for(int i=0;i<numRounds;i++){
      if (i==numRounds-1){
        lastRound = true;
      }
      currentTime = System.currentTimeMillis();
      timeout = currentTime+receivingTimeout;
      resetActions();
      while(System.currentTimeMillis()<timeout){
        Thread.sleep(20);
        getActions();
        if (playerAborted()){
          throw new IOException("Player cleanly aborted:"+Array.arrayToNakedString(observedActions));
        }
        if (actionsComplete()){
          break;
        }
      }
      
      if (playerAborted()){
        throw new IOException("Player cleanly aborted");
      }
      
      if (!actionsComplete()){
        throw new IOException("Agent timed out");
      }
      
      if (!areActionsLegal()){
        throw new IOException("Player cleanly aborted");
      }
      
      
      sendOutcomes();
    }
    removeAllUsers();
    recordInLog("Game completed");
    closeLog();
    } catch (Exception io){
      io.printStackTrace();
      abortGame();
    } 
    
    gameInProgress=false;

  }
  @Override
  public void handleChat(String chat) {
    for(int i=0;i<g.getNumPlayers();i++){
      Agent agent = this.getUser(i);
      if (agent!=null){
        try{
          agent.sendMessage(chat);
        } catch (IOException io){
          agent.gameOver = true;
        }
      }
    }
    
    
  }
  
}
