package net;

import java.io.IOException;
import java.io.InputStream;

/**
 * A class that stores partially received messages, allowing users to peek or receive messages.
 * 
 * If a chat message is received, it should probably be handled by a separate thread.
 * @author maz
 *
 */
public class InputStreamListener {
  /**
   * A response that is being built up.
   */
  private String partialResponse;
  private InputStream is;
  private String messageTerminator;
  private ChatListener chatListener;
  private boolean interrupted;
  
  public InputStreamListener(InputStream is, String messageTerminator){
    partialResponse = "";
    this.is = is;
    this.messageTerminator = messageTerminator;
    interrupted = false;
  }
  
  public void addChatListener(ChatListener listener){
    this.chatListener = listener;
  }
  /**
   * Test if the message is complete (contains a terminal character)
   */
  public boolean isComplete(String result) {
    return result.endsWith(messageTerminator);
  }
  
  
  
  
  
  public synchronized String peekAtMessage() throws IOException,
      InterruptedException {
    while (true) {
      while ((!isComplete(partialResponse)) && (!interrupted)
          && is.available() >= 1) {
        int result = is.read();
        if (result == -1) {
          throw new IOException("End of file");
        }
        char c = (char) (result);
        // System.out.println("READ:"+(int)c);
        partialResponse = partialResponse + c;
      }
      if (isComplete(partialResponse)) {
        if (partialResponse.startsWith("CHAT")) {
          if (chatListener != null) {
            chatListener.handleChat(chop(partialResponse));
          }
          partialResponse = "";
        } else {
          return chop(partialResponse);
        }
      } else {
        return null;
      }
    }
  }
  
  /**
   * Grabs the last message sent, and clears the buffer.
   * @return
   */
  public synchronized String grabMessage(){
    String result = partialResponse;
    partialResponse = "";
    if (isComplete(result)){
      return chop(result);
    }
    return null;
  }
 
  
  public String chop(String response){
    return response.substring(0,response.length()-messageTerminator.length());
  }
  /**
   * Blocks until a line is ready to be read.
   * If two threads call receiveMessage simultaneously on the same InputStreamListener object, 
   * it is possible for one of the threads to get a null pointer. 
   * 
   * @return
   * @throws IOException
   * @throws InterruptedException
   */
  public String receiveMessage() throws IOException, InterruptedException{
    while(peekAtMessage()==null){
      Thread.sleep(20);
    }
    return grabMessage();
  }
  
  public void close() throws IOException{
    is.close();
  }
}
